from __future__ import annotations

from collections.abc import Iterable

from satori.client.account import Account
from satori.element import Element
from satori.model import Channel, Event, Member, MessageObject, PageResult, Role, User


class ContextSession:
    """在 Satori-Python 的 Session 的基础上存储了当次事件的信息"""

    def __init__(self, account: Account, event: Event):
        self.account = account
        self.context = event

    def __getattr__(self, item):
        return getattr(self.account.session, item)

    async def send(
        self,
        message: str | Iterable[str | Element],
    ) -> list[MessageObject]:
        return await self.account.session.send(self.context, message)

    async def send_message(
        self,
        message: str | Iterable[str | Element],
    ) -> list[MessageObject]:
        """发送消息

        参数:
            message: 要发送的消息
        """
        if not self.context.channel:
            raise RuntimeError("Event cannot be replied to!")
        return await self.account.session.send_message(self.context.channel, message)

    async def send_private_message(
        self,
        message: str | Iterable[str | Element],
    ) -> list[MessageObject]:
        """发送私聊消息

        参数:
            message: 要发送的消息
        """
        if not self.context.user:
            raise RuntimeError("Event cannot be replied to!")
        return await self.account.session.send_private_message(self.context.user, message)

    async def update_message(
        self,
        message: str | Iterable[str | Element],
    ):
        """更新消息

        参数:
            message: 要更新的消息
        """
        if not self.context.channel:
            raise RuntimeError("Event cannot be replied to!")
        if not self.context.message:
            raise RuntimeError("Event cannot update message")
        return await self.account.session.update_message(
            self.context.channel, self.context.message.id, message
        )

    async def message_create(
        self,
        content: str,
    ) -> list[MessageObject]:
        if not self.context.channel:
            raise RuntimeError("Event cannot be replied to!")
        return await self.account.session.message_create(self.context.channel.id, content)

    async def message_delete(self) -> None:
        if not self.context.channel:
            raise RuntimeError("Event cannot be replied to!")
        if not self.context.message:
            raise RuntimeError("Event cannot update message")
        await self.account.session.message_delete(
            self.context.channel.id,
            self.context.message.id,
        )

    async def message_update(
        self,
        content: str,
    ) -> None:
        if not self.context.channel:
            raise RuntimeError("Event cannot be replied to!")
        if not self.context.message:
            raise RuntimeError("Event cannot update message")
        await self.account.session.message_update(
            self.context.channel.id,
            self.context.message.id,
            content,
        )

    async def channel_create(self, data: Channel) -> Channel:
        if not self.context.guild:
            raise RuntimeError("Event cannot use to create channel!")
        return await self.account.session.channel_create(self.context.guild.id, data)

    async def channel_list(self, next_token: str | None = None) -> PageResult[Channel]:
        if not self.context.guild:
            raise RuntimeError("Event cannot use to list channel!")
        return await self.account.session.channel_list(self.context.guild.id, next_token)

    async def channel_update(
        self,
        data: Channel,
    ) -> None:
        if not self.context.channel:
            raise RuntimeError("Event cannot use to update channel!")
        return await self.account.session.channel_update(self.context.channel.id, data)

    async def channel_delete(self) -> None:
        if not self.context.channel:
            raise RuntimeError("Event cannot use to delete channel!")
        return await self.account.session.channel_delete(
            self.context.channel.id,
        )

    async def user_channel_create(self) -> Channel:
        if not self.context.user:
            raise RuntimeError("Event cannot use to create user channel!")
        return await self.account.session.user_channel_create(
            self.context.user.id, self.context.guild.id if self.context.guild else None
        )

    async def guild_member_list(self, next_token: str | None = None) -> PageResult[Member]:
        if not self.context.guild:
            raise RuntimeError("Event cannot use to list member!")
        return await self.account.session.guild_member_list(self.context.guild.id, next_token)

    async def guild_member_get(self, user_id: str | None = None) -> Member:
        if not self.context.guild:
            raise RuntimeError("Event cannot use to get member!")
        if user_id:
            return await self.account.session.guild_member_get(self.context.guild.id, user_id)
        if not self.context.user:
            raise RuntimeError("Event cannot use to get member!")
        return await self.account.session.guild_member_get(self.context.guild.id, self.context.user.id)

    async def guild_member_kick(self, user_id: str | None = None, permanent: bool = False) -> None:
        if not self.context.guild:
            raise RuntimeError("Event cannot use to kick member!")
        if user_id:
            return await self.account.session.guild_member_kick(self.context.guild.id, user_id, permanent)
        if not self.context.user:
            raise RuntimeError("Event cannot use to kick member!")
        return await self.account.session.guild_member_kick(
            self.context.guild.id, self.context.user.id, permanent
        )

    async def guild_member_role_set(self, role_id: str, user_id: str | None = None) -> None:
        if not self.context.guild:
            raise RuntimeError("Event cannot use to guild member role set!")
        if user_id:
            return await self.account.session.guild_member_role_set(self.context.guild.id, user_id, role_id)
        if not self.context.user:
            raise RuntimeError("Event cannot use to guild member role set!")
        return await self.account.session.guild_member_role_set(
            self.context.guild.id, self.context.user.id, role_id
        )

    async def guild_member_role_unset(self, role_id: str, user_id: str | None = None) -> None:
        if not self.context.guild:
            raise RuntimeError("Event cannot use to guild member role unset!")
        if user_id:
            return await self.account.session.guild_member_role_unset(self.context.guild.id, user_id, role_id)
        if not self.context.user:
            raise RuntimeError("Event cannot use to guild member role unset!")
        return await self.account.session.guild_member_role_unset(
            self.context.guild.id, self.context.user.id, role_id
        )

    async def guild_role_list(self, next_token: str | None = None) -> PageResult[Role]:
        if not self.context.guild:
            raise RuntimeError("Event cannot use to list role!")
        return await self.account.session.guild_role_list(self.context.guild.id, next_token)

    async def guild_role_create(
        self,
        role: Role,
    ) -> Role:
        if not self.context.guild:
            raise RuntimeError("Event cannot use to create role!")
        return await self.account.session.guild_role_create(self.context.guild.id, role)

    async def guild_role_update(
        self,
        role_id: str,
        role: Role,
    ) -> None:
        if not self.context.guild:
            raise RuntimeError("Event cannot use to update role!")
        return await self.account.session.guild_role_update(self.context.guild.id, role_id, role)

    async def guild_role_delete(self, role_id: str) -> None:
        if not self.context.guild:
            raise RuntimeError("Event cannot use to delete role!")
        return await self.account.session.guild_role_delete(
            self.context.guild.id,
            role_id,
        )

    async def reaction_create(
        self,
        emoji: str,
    ) -> None:
        if not self.context.channel:
            raise RuntimeError("Event cannot be replied to!")
        if not self.context.message:
            raise RuntimeError("Event cannot create reaction")
        return await self.account.session.reaction_create(
            self.context.channel.id, self.context.message.id, emoji
        )

    async def reaction_delete(
        self,
        emoji: str,
        user_id: str | None = None,
    ) -> None:
        if not self.context.channel:
            raise RuntimeError("Event cannot be replied to!")
        if not self.context.message:
            raise RuntimeError("Event cannot delete reaction")
        return await self.account.session.reaction_delete(
            self.context.channel.id, self.context.message.id, emoji, user_id
        )

    async def reaction_clear(
        self,
        emoji: str | None = None,
    ) -> None:
        if not self.context.channel:
            raise RuntimeError("Event cannot be replied to!")
        if not self.context.message:
            raise RuntimeError("Event cannot clear reaction")
        return await self.account.session.reaction_clear(
            self.context.channel.id,
            self.context.message.id,
            emoji,
        )

    async def reaction_list(
        self,
        emoji: str,
        next_token: str | None = None,
    ) -> PageResult[User]:
        if not self.context.channel:
            raise RuntimeError("Event cannot be replied to!")
        if not self.context.message:
            raise RuntimeError("Event cannot list reaction")
        return await self.account.session.reaction_list(
            self.context.channel.id, self.context.message.id, emoji, next_token
        )



from __future__ import annotations

import inspect
from contextlib import suppress
from dataclasses import dataclass, field
from os import PathLike
from pathlib import Path
from typing import Callable
from loguru import logger
import importlib

from arclet.letoderea import BaseEvent, Publisher, system_ctx

dispatchers = {}


class PluginDispatcher(Publisher):
    def __init__(
        self,
        plugin: "Plugin",
        *events: type[BaseEvent],
        predicate: Callable[[BaseEvent], bool] | None = None,
    ):
        super().__init__(plugin.name, *events, predicate=predicate)  # type: ignore
        self.plugin = plugin
        if es := system_ctx.get():
            es.register(self)
        else:
            dispatchers[self.id] = self

    on = Publisher.register
    handle = Publisher.register


@dataclass
class Plugin:
    author: list[str] = field(default_factory=list)
    name: str | None = None
    version: str | None = None
    license: str | None = None
    urls: dict[str, str] | None = None
    description: str | None = None
    icon: str | None = None
    classifier: list[str] = field(default_factory=list)
    dependencies: list[str] = field(default_factory=list)

    standards: list[str] = field(default_factory=list)
    frameworks: list[str] = field(default_factory=list)
    config_endpoints: list[str] = field(default_factory=list)
    component_endpoints: list[str] = field(default_factory=list)

    _dispatchers: dict[str, PluginDispatcher] = field(default_factory=dict, init=False)

    def __post_init__(self):
        self.name = self.name or self.__module__

    def dispatch(self, *events: type[BaseEvent], predicate: Callable[[BaseEvent], bool] | None = None):
        disp = PluginDispatcher(self, *events, predicate=predicate)
        self._dispatchers[disp.id] = disp
        return disp

    def dispose(self):
        for disp in self._dispatchers.values():
            if disp.id in dispatchers:
                del dispatchers[disp.id]
            if es := system_ctx.get():
                es.publishers.pop(disp.id, None)
        self._dispatchers.clear()


def load_plugin(path: str) -> list[Plugin] | None:
    """
    以导入路径方式加载模块

    Args:
        path (str): 模块路径
    """
    with suppress(ModuleNotFoundError):
        imported_module = importlib.import_module(path, path)
        logger.success(f"loaded plugin {path!r}")
        return [
            m
            for _, m in inspect.getmembers(
                imported_module, lambda x: isinstance(x, Plugin)
            )
        ]
    logger.warning(f"failed to load plugin {path!r}")


def load_plugins(dir_: str | PathLike | Path):
    path = dir_ if isinstance(dir_, Path) else Path(dir_)
    if path.is_dir():
        for p in path.iterdir():
            if p.suffix in (".py", "") and p.stem not in {"__init__", "__pycache__"}:
                load_plugin(".".join(p.parts[:-1:1]) + "." + p.stem)
    elif path.is_file():
        load_plugin(".".join(path.parts[:-1:1]) + "." + path.stem)



from __future__ import annotations

from dataclasses import dataclass, field, fields
from datetime import datetime
from typing import ClassVar

from arclet.letoderea import Contexts, Param, Provider
from satori import ArgvInteraction, ButtonInteraction, Channel
from satori import Event as SatoriEvent
from satori import EventType, Guild, Login, Member, Quote, Role, User
from satori.client import Account
from satori.model import MessageObject
from tarina import gen_subclass

from .message import MessageChain


@dataclass
class Event:
    type: ClassVar[EventType]
    _origin: SatoriEvent = field(init=False)

    id: int
    timestamp: datetime
    account: Account

    @classmethod
    def parse(cls, account: Account, origin: SatoriEvent):
        fs = fields(cls)
        attrs = {"account": account}
        for fd in fs:
            if not fd.init:
                continue
            if attr := getattr(origin, fd.name, None):
                attrs[fd.name] = attr
        res = cls(**attrs)
        res._origin = origin
        return res

    async def gather(self, context: Contexts):
        context["$account"] = self.account
        context["$origin_event"] = self._origin

    class AccountProvider(Provider[Account]):
        async def __call__(self, context: Contexts):
            return context["$account"]

    class TimeProvider(Provider[datetime]):
        async def __call__(self, context: Contexts):
            return context["$event"].timestamp

    class OperatorProvider(Provider[User]):
        priority = 10

        def validate(self, param: Param):
            return param.name == "operator" and super().validate(param)

        async def __call__(self, context: Contexts):
            return context["$origin_event"].operator

    class UserProvider(Provider[User]):
        async def __call__(self, context: Contexts):
            return context["$origin_event"].user

    class MessageProvider(Provider[MessageObject]):
        async def __call__(self, context: Contexts):
            return context["$origin_event"].message

    class ChannelProvider(Provider[Channel]):
        async def __call__(self, context: Contexts):
            return context["$origin_event"].channel

    class GuildProvider(Provider[Guild]):
        async def __call__(self, context: Contexts):
            return context["$origin_event"].guild


class NoticeEvent(Event):
    pass


@dataclass
class FriendEvent(NoticeEvent):
    user: User

    async def gather(self, context: Contexts):
        await super().gather(context)
        context["user"] = self.user


class FriendRequestEvent(FriendEvent):
    type = EventType.FRIEND_REQUEST


@dataclass
class GuildEvent(NoticeEvent):
    guild: Guild

    async def gather(self, context: Contexts):
        await super().gather(context)
        context["guild"] = self.guild


class GuildAddedEvent(GuildEvent):
    type = EventType.GUILD_ADDED


class GuildRemovedEvent(GuildEvent):
    type = EventType.GUILD_REMOVED


class GuildRequestEvent(GuildEvent):
    type = EventType.GUILD_REQUEST


class GuildUpdatedEvent(GuildEvent):
    type = EventType.GUILD_UPDATED


@dataclass
class GuildMemberEvent(GuildEvent):
    user: User
    member: Member | None = None

    async def gather(self, context: Contexts):
        await super().gather(context)
        context["user"] = self.user
        if self.member:
            context["member"] = self.member


class GuildMemberAddedEvent(GuildMemberEvent):
    type = EventType.GUILD_MEMBER_ADDED


class GuildMemberRemovedEvent(GuildMemberEvent):
    type = EventType.GUILD_MEMBER_REMOVED


class GuildMemberRequestEvent(GuildMemberEvent):
    type = EventType.GUILD_MEMBER_REQUEST


class GuildMemberUpdatedEvent(GuildMemberEvent):
    type = EventType.GUILD_MEMBER_UPDATED


@dataclass
class GuildRoleEvent(GuildEvent):
    role: Role

    async def gather(self, context: Contexts):
        await super().gather(context)
        context["role"] = self.role

    class RoleProvider(Provider[Role]):
        async def __call__(self, context: Contexts):
            return context.get("role")


class GuildRoleCreatedEvent(GuildRoleEvent):
    type = EventType.GUILD_ROLE_CREATED


class GuildRoleDeletedEvent(GuildRoleEvent):
    type = EventType.GUILD_ROLE_DELETED


class GuildRoleUpdatedEvent(GuildRoleEvent):
    type = EventType.GUILD_ROLE_UPDATED


@dataclass
class LoginEvent(NoticeEvent):
    login: Login

    async def gather(self, context: Contexts):
        await super().gather(context)
        context["login"] = self.login

    class LoginProvider(Provider[Login]):
        async def __call__(self, context: Contexts):
            return context.get("login")


class LoginAddedEvent(LoginEvent):
    type = EventType.LOGIN_ADDED


class LoginRemovedEvent(LoginEvent):
    type = EventType.LOGIN_REMOVED


class LoginUpdatedEvent(LoginEvent):
    type = EventType.LOGIN_UPDATED


class MessageContentProvider(Provider[MessageChain]):
    async def __call__(self, context: Contexts):
        return context["$message_content"]


class QuoteProvider(Provider[Quote]):
    async def __call__(self, context: Contexts):
        return context["$event"].quote


@dataclass
class MessageEvent(Event):
    channel: Channel
    user: User
    message: MessageObject
    content: MessageChain = field(init=False)
    guild: Guild | None = None
    member: Member | None = None
    quote: Quote | None = None

    providers = [MessageContentProvider, QuoteProvider]

    def __post_init__(self):
        self.content = MessageChain(self.message.message)
        if self.content.has(Quote):
            self.quote = self.content.get(Quote, 1)[0]
            self.content = self.content.exclude(Quote)

    async def gather(self, context: Contexts):
        await super().gather(context)
        context["user"] = self.user
        context["channel"] = self.channel
        context["$message_origin"] = self.message
        context["$message_content"] = self.content
        if self.member:
            context["member"] = self.member
        if self.guild:
            context["guild"] = self.guild


class MessageCreatedEvent(MessageEvent):
    type = EventType.MESSAGE_CREATED


class MessageDeletedEvent(MessageEvent):
    type = EventType.MESSAGE_DELETED


class MessageUpdatedEvent(MessageEvent):
    type = EventType.MESSAGE_UPDATED


@dataclass
class ReactionEvent(NoticeEvent):
    channel: Channel
    user: User
    message: MessageObject
    content: MessageChain = field(init=False)
    quote: Quote | None = None
    guild: Guild | None = None
    member: Member | None = None

    providers = [MessageContentProvider, QuoteProvider]

    def __post_init__(self):
        self.content = MessageChain(self.message.message)
        if self.content.has(Quote):
            self.quote = self.content.get(Quote, 1)[0]
            self.content = self.content.exclude(Quote)

    async def gather(self, context: Contexts):
        await super().gather(context)
        context["user"] = self.user
        context["channel"] = self.channel
        context["$message_origin"] = self.message
        context["$message_content"] = self.content
        if self.member:
            context["member"] = self.member
        if self.guild:
            context["guild"] = self.guild


class ReactionAddedEvent(ReactionEvent):
    type = EventType.REACTION_ADDED


class ReactionRemovedEvent(ReactionEvent):
    type = EventType.REACTION_REMOVED


class InternalEvent(Event):
    type = EventType.INTERNAL


class InteractionEvent(NoticeEvent):
    pass


@dataclass
class InteractionButtonEvent(InteractionEvent):
    type = EventType.INTERACTION_BUTTON

    button: ButtonInteraction

    async def gather(self, context: Contexts):
        await super().gather(context)
        context["button"] = self.button

    class ButtonProvider(Provider[ButtonInteraction]):
        async def __call__(self, context: Contexts):
            return context.get("button")


class InteractionCommandEvent(InteractionEvent):
    type = EventType.INTERACTION_COMMAND


@dataclass
class InteractionCommandArgvEvent(InteractionCommandEvent):
    argv: ArgvInteraction

    async def gather(self, context: Contexts):
        await super().gather(context)
        context["argv"] = self.argv

    class ArgvProvider(Provider[ArgvInteraction]):
        async def __call__(self, context: Contexts):
            return context.get("argv")


@dataclass
class InteractionCommandMessageEvent(InteractionCommandEvent):
    channel: Channel
    user: User
    message: MessageObject
    content: MessageChain = field(init=False)
    quote: Quote | None = None
    guild: Guild | None = None
    member: Member | None = None

    providers = [MessageContentProvider, QuoteProvider]

    def __post_init__(self):
        self.content = MessageChain(self.message.message)
        if self.content.has(Quote):
            self.quote = self.content.get(Quote, 1)[0]
            self.content = self.content.exclude(Quote)

    async def gather(self, context: Contexts):
        await super().gather(context)
        context["user"] = self.user
        context["channel"] = self.channel
        context["$message_origin"] = self.message
        context["$message_content"] = self.content
        if self.member:
            context["member"] = self.member
        if self.guild:
            context["guild"] = self.guild


MAPPING = {}

for cls in gen_subclass(Event):
    if hasattr(cls, "type"):
        MAPPING[cls.type.value] = cls


def event_parse(account: Account, event: SatoriEvent):
    try:
        return MAPPING[event.type].parse(account, event)
    except KeyError:
        raise NotImplementedError from None
